open Utils
open Logic_ast
open Comp_ast
open Logic_core
open Logic_typing

let case_kind f1 f2 f3 ~case:(_,t1) = 
  match t1 with
      CHol -> f1
    | CCtx -> f2
    | _ -> f3

let flatten_all_cterm =
  cterm_map ~cholterm:(fun t -> CHolTerm(flatten_all_lmodal t))
            ~cctxterm:(fun t -> CCtxTerm(flatten_all_lctxdesc t))

let subst_bound_ctx subst term =
  let term' = CtxbindCtermS.subst_bound subst term in
  flatten_all_cterm term'

let subst_bound ~case:t1 ~term:t2 ~subst:t3 =
  case_kind
    (function CHolTerm(t) -> MetabindCtermS.subst_bound t | _ -> failwith "problem in subst_bound")
    (function CCtxTerm(t) -> subst_bound_ctx t | _ -> failwith "problem in subst_bound")
    CbindCtermS.subst_bound
    ~case:t1
    t3
    t2

let get_cdef_term n cdefenv = ExtDict.getsnd n cdefenv
let get_cdef_type n cdefenv = ExtDict.getfst n cdefenv      

let flatten_closure (metasubst,ctxsubst,compsubst) e =
  flatten_all_cterm
    (MetabindCtermS.subst_bound_list (List.rev metasubst)
       (CtxbindCtermS.subst_bound_list (List.rev ctxsubst)
	  (CbindCtermS.subst_bound_list (List.rev compsubst) e)))
    

let rec ctype_whnf e =
  match e with
      CApp(t1, t3) ->
	(match ctype_whnf t1 with
	     CLambda(s, t1, t2) ->
	       ctype_whnf (subst_bound ~case:s ~term:t2 ~subst:t3)
	   | t1' -> CApp(t1', t3))
    | CAny(l) ->
	(match !l with Some(t) -> ctype_whnf t | None -> e)
    | CClosure(subst, e) ->
	ctype_whnf (flatten_closure subst e)
    | t -> t

let rec ctype_whdelta cdefenv e =
  monadic option_monad in begin
    let rec aux e =
      match e with
	  CNVar(n) -> cmd { return (get_cdef_term n cdefenv) }
	| CApp(e1, e2) ->  cmd { e1' <- aux e1 then
				 return (CApp(e1', e2)) }
	| e -> None
    in
      cmd { e' <- aux e then
	    return (ctype_whnf e') }
  end

let rec ctype_fullwhdelta cdefenv e =
  let e' = ctype_whnf e in
    match ctype_whdelta cdefenv e with
	None -> e'
      | Some e'' -> ctype_fullwhdelta cdefenv e''


let ctype_equal defenv cdefenv e1 e2 =
  let rec aux e1 e2 = 
    match ctype_whnf e1, ctype_whnf e2 with
	CSort(s1), CSort(s2) -> s1 = s2
      | CPi(_, a1, b1), CPi(_, a2, b2) -> aux a1 a2 && aux b1 b2
      | CLambda(_, a1, b1), CLambda(_, a2, b2) -> aux a1 a2 && aux b1 b2
      | CApp(a1, b1), CApp(a2, b2) when aux a1 a2 && aux b1 b2 -> true
      | CHolTerm(t1), CHolTerm(t2) -> lmodal_equal defenv t1 t2
      | CCtxTerm(t1), CCtxTerm(t2) -> lctxdesc_equal defenv t1 t2
      | CUnitType, CUnitType -> true
      | CSigma(_, a1, b1), CSigma(_, a2, b2) -> aux a1 a2 && aux b1 b2
      | CBVar(i1), CBVar(i2) -> i1 = i2
      | CFVar(i1), CFVar(i2) -> i1 = i2
      | CProdType(a1,b1), CProdType(a2,b2) -> aux a1 a2 && aux b1 b2
      | CRecType(_,a1,b1), CRecType(_,a2,b2) -> aux a1 a2 && aux b1 b2
      | CSumType(b1), CSumType(b2) -> List.for_all2 aux b1 b2
      | CRefType(t1), CRefType(t2) -> aux t1 t2
      | CIntType, CIntType -> true
      | CBoolType, CBoolType -> true
      | CArrayType(t1), CArrayType(t2) -> aux t1 t2
      | CAny(l), e2 ->
	  (match !l with
	       Some(e1) -> aux e1 e2
	     | None -> l := Some e2; true)
      | e1, CAny(l) ->
	  (match !l with
	       Some(e2) -> aux e1 e2
	     | None -> l := Some e1; true)
      | CNVar(s1), CNVar(s2) when s1 = s2 -> true
      | CNVar(s1), e2 when aux (get_cdef_term s1 cdefenv) e2 -> true
      | e1, CNVar(s2) when aux e1 (get_cdef_term s2 cdefenv) -> true
      | e1, e2 -> (match ctype_whdelta cdefenv e1, ctype_whdelta cdefenv e2 with
		       Some(e1), Some(e2) -> aux e1 e2
		     | Some(e1), None -> aux e1 e2
		     | None, Some(e2) -> aux e1 e2
		     | None, None -> false)
  in
    aux e1 e2

(* SPECIFICATION OF THE LANGUAGE *)
let ctype_spec_abstractions =
  [ (CType, CType);
    (CCtx,  CType);
    (CHol,  CType);
    (CKind, CType);
    (CKind, CKind);
    (CHol,  CKind);
    (CCtx,  CKind)]

let ctype_spec_existentials =
  [ (CHol, CType);
    (CCtx, CType) ]

let add_to_metaenv tp (defenv, cdefenv, metaenv, ctxenv, compenv) =
  let metaenv' = tp :: metaenv in
  let metaenv'' = List.map (MetabindModal.bumpup_fvar 1) metaenv' in
  let compenv'' = List.map (MetabindCterm.bumpup_fvar 1) compenv in
    (defenv, cdefenv, metaenv'', ctxenv, compenv'')

let add_to_ctxenv tp (defenv, cdefenv, metaenv, ctxenv, compenv) =
  let ctxenv' = () :: ctxenv in
  let metaenv'' = List.map (CtxbindModal.bumpup_fvar 1) metaenv in
  let compenv'' = List.map (CtxbindCterm.bumpup_fvar 1) compenv in
    (defenv, cdefenv, metaenv'', ctxenv', compenv'')

let add_to_compenv tp (defenv, cdefenv, metaenv, ctxenv, compenv) =
  let compenv' = tp :: compenv in
  let compenv'' = List.map (CbindCterm.bumpup_fvar 1) compenv' in
    (defenv, cdefenv, metaenv, ctxenv, compenv'')

let add_to_env =
  case_kind
    (function CHolTerm(t) -> add_to_metaenv t | _ -> failwith "add_to_env with HOL binder and non-HOL term")
    (function CSort(CCtx) -> add_to_ctxenv () | _ -> failwith "add_to_env with CTX binder and non-CTX term")
    (function CHolTerm(_) | CSort(CCtx) -> failwith "add_to_env with COMP binder and non-COMP term" | t -> add_to_compenv t)
let open_up    = case_kind MetabindCterm.open_up CtxbindCterm.open_up CbindCterm.open_up
let close_down = case_kind MetabindCterm.close_down CtxbindCterm.close_down CbindCterm.close_down
let has_bound_var = case_kind MetabindCterm.has_bound_var CtxbindCterm.has_bound_var CbindCterm.has_bound_var

let open_up_2 ~case1:((_,var1) as s1) ~case2:((_,var2) as s2) =
  if var1 = var2 then open_up ~case:s1 ~howmany:2 else compose (open_up ~case:s1) (open_up ~case:s2)

let close_down_2 ~case1:((_,var1) as s1) ~case2:((_,var2) as s2) =
  if var1 = var2 then close_down ~case:s1 ~howmany:2 else compose (close_down ~case:s1) (close_down ~case:s2)



let rec type_of_cterm ((defenv, cdefenv, metaenv, ctxenv, compenv) as env) (e : cterm) =
  match e with
      CSort(CType) -> CSort(CKind)
    | CSort(CKind) -> failwith "Kind does not have a classifier"
    | CSort(CCtx)  -> failwith "ctx is not part of the cterm language"
    | CSort(CHol)  -> failwith "HOL is not part of the cterm language"
    | CPi(var, t1, t2) ->
	(let t1_s = csort_of ~case:var env t1 in
	 let env' = add_to_env ~case:var t1 env in
	 let t2'  = open_up ~case:var t2 in
	 let t2_s = csort_of_cterm env' t2' in
	   if List.mem (t1_s, t2_s) ctype_spec_abstractions then
	     CSort(t2_s)
	   else
	     failwith "non-allowed abstraction used")
    | CLambda(var, t1, e2) ->
	(let env' = add_to_env ~case:var t1 env in
	 let t2   = type_of_cterm env' (open_up ~case:var e2) in
	 let tp   = CPi(var, t1, close_down ~case:var t2) in
	 let _    = type_of_cterm env tp in
	   tp)
    | CApp(e1, e2) ->
	(let t1 = ctype_fullwhdelta cdefenv (type_of_cterm env e1) in
	   match t1 with
	       CPi(var, t, t') ->
		 let t2 = type_of ~case:var env e2 in
		   if ctype_equal defenv cdefenv t2 t then
		     subst_bound ~case:var ~term:t' ~subst:e2
		   else
		     failwith "types don't match in application"
	     | _ -> failwith "applying something to a non-functional type")
    | CHolTerm(t) ->
	failwith "HOL terms are not part of the cterm language"
    | CCtxTerm(t) ->
	failwith "ctx terms are not part of the cterm language"
    | CUnitType -> CSort(CType)
    | CUnitExpr -> CUnitType
    | CSigma(var, t1, t2) ->
	(let t1_s = csort_of ~case:var env t1 in
	 let env' = add_to_env ~case:var t1 env in
	 let t2'  = open_up ~case:var t2 in
	 let t2_s = csort_of_cterm env' t2' in
	   if List.mem (t1_s, t2_s) ctype_spec_existentials then
	     CSort(t2_s)
	   else
	     failwith "non-allowed existential type used")
    | CPack(t1, var, ret, e) ->
	(let t1_t = type_of ~case:var env t1 in
	 let tp = CSigma(var, t1_t, ret) in
	 let _ = type_of_cterm env tp in
	 let e_t = type_of_cterm env e in
	 let e_expected = subst_bound ~case:var ~term:ret ~subst:t1 in
	   if ctype_equal defenv cdefenv e_t e_expected then
	     tp
	   else
	     failwith "body of existential doesn't have expected type")
    | CUnpack(t1, var1, var2, e) ->
	(let t1_t = ctype_fullwhdelta cdefenv (type_of_cterm env t1) in
	   match t1_t with
	       CSigma(_, tt1, tt2) ->
		 let env' = add_to_env ~case:var1 tt1 env in
		 let tt2' = open_up ~case:var1 tt2 in
		 let env'' = add_to_env ~case:var2 tt2' env' in
		 let e' = open_up_2 ~case1:var1 ~case2:var2 e in
		 let e_t = close_down_2 ~case1:var1 ~case2:var2 (type_of_cterm env'' e') in
		   if has_bound_var ~case:var1 0 e_t then
		     failwith "witness should not escape from existential unpacking"
		   else if csort_of_cterm env e_t = CType then
		     e_t
		   else
		     failwith "unpacking inside non-computational term"
	     | _ -> failwith "unpacking a non-existential term")
    | CBVar(i) -> failwith "type_of_cterm on a bound variable!"
    | CFVar(i) -> List.nth compenv i
    | CNVar(s) -> get_cdef_type s cdefenv
    | CAny(l) -> (match !l with Some(e) -> type_of_cterm env e | None -> CAny(ref None))
    | CHolCase(t,var,ret,branches) ->
	(let t_t = type_of_HOL env t in
	 let env' = add_to_env ~case:var t_t env in
	 let ret' = open_up ~case:var ret in
	 let ret_s = csort_of_cterm env' ret' in
	 let ret_f t = subst_bound ~case:var ~term:ret ~subst:t in
	   if ret_s = CType then
	     (let _ = List.iter (check_branch env t_t ret_f) branches in
		ret_f t)
	   else
	     failwith "matching on HOL terms is only allowed for computational terms")
    | CProdType(t1,t2) ->
	(if ctype_equal defenv cdefenv (type_of_cterm env t1) (CSort(CType)) &&
	    ctype_equal defenv cdefenv (type_of_cterm env t2) (CSort(CType)) then
	   CSort(CType)
	 else
	   failwith "product type with non-product terms")
    | CTuple(e1,e2) ->
	(let t1 = type_of_cterm env e1 in
	 let t2 = type_of_cterm env e2 in
	 let tp = CProdType(t1,t2) in
	 let _ = type_of_cterm env tp in
	   tp)
    | CProj(i,e) ->
	(let t = ctype_fullwhdelta cdefenv (type_of_cterm env e) in
	   match t with
	       CProdType(t1,t2) ->
		 (if i = 1 then t1
		  else if i = 2 then t2
		  else failwith "proj with invalid index")
	     | _ -> failwith "proj on non-tuple")
    | CRecType(var, k, t) ->
	(let _ = if csort_of_cterm env k = CKind then () else failwith "recursive type's kind is not CKind" in
	 let env' = add_to_env ~case:var k env in
	 let t' = open_up ~case:var t in
	   if ctype_equal defenv cdefenv (type_of_cterm env' t') k then
	     k
	   else
	     failwith "recursive type's definition does not match expected kind")
    | CFold(e1, t2) ->
	(let t1 = type_of_cterm env e1 in
	 let _ = type_of_cterm env t2 in
	   match ctype_fullwhdelta cdefenv t2 with
	     CRecType(var, k, t) ->
	       (let t1exp = subst_bound ~case:var ~subst:t2 ~term:t in
		  if ctype_equal defenv cdefenv t1 t1exp then
		    t2
		  else
		    failwith "fold applied to invalid instantiation of recursive type")
	     | _ ->
		 failwith "fold with non-recursive type")
    | CUnfold(e1) ->
	(let t1 = type_of_cterm env e1 in
	   match ctype_fullwhdelta cdefenv t1 with
	       CRecType(var, k, t) ->
		 subst_bound ~case:var ~subst:t1 ~term:t
	     | _ ->
		 failwith "unfold with non-recursive type argument")
    | CSumType(branches) ->
	(let checkbranch branch = ctype_equal defenv cdefenv (type_of_cterm env branch) (CSort CType) in
	   if List.for_all checkbranch branches then
	     CSort(CType)
	   else
	     failwith "there is a branch with non-computational type")
    | CCtor(i,t,par) ->
	(let t' = ctype_fullwhdelta cdefenv t in
	 let par_t = type_of_cterm env par in
	   match t' with
	       CSumType(branches) ->
		 let branch_t = List.nth branches i in
		   if ctype_equal defenv cdefenv par_t branch_t then 
		     t
		   else
		     failwith "parameter of constructor is not of right type"
	     | _ -> failwith "constructor used with non-sum type")
    | CMatch(e,branches) ->
	(let e_t = type_of_cterm env e in
	 let constrs_t = match ctype_fullwhdelta cdefenv e_t with CSumType(b) -> b | _ -> failwith "match on term of non-sum type" in
	   (* this is a hack. take care of it *)
	 let t = 
	   let constr_t = List.nth constrs_t 1 in
	   let (var,body) = List.nth branches 1 in
	   let env' = add_to_env ~case:var constr_t env in
	   let body' = open_up ~case:var body in
	   let body_t = close_down ~case:var (type_of_cterm env' body') in
	     body_t
	 in
	 let checkbranch (var,body) constr_t =
	   let env' = add_to_env ~case:var constr_t env in
	   let body' = open_up ~case:var body in
	   let body_t = close_down ~case:var (type_of_cterm env' body') in
	     if not (ctype_equal defenv cdefenv body_t t) then failwith "branch doesn't match expected type"
	 in
	 let _ = List.iter2 checkbranch branches constrs_t in
	 let _ = if not (ctype_equal defenv cdefenv (type_of_cterm env t) (CSort(CType))) then failwith "match with non-computational type as result" in
	   t)
    | CLetRec(defs, e) ->
	(let n = List.length defs in
	 let check_def_type (_,t,_) = if not (ctype_equal defenv cdefenv (type_of_cterm env t) (CSort(CType))) then failwith "letrec with non-computational type of definition" in
	 let _ = List.iter check_def_type defs in
	 let env' = List.fold_left (fun env (var,t,_) -> add_to_env ~case:var t env) env defs in
	 let check_def (_,t,e) =
	   let e' = CbindCterm.open_up ~howmany:n e in
	   let e_t = CbindCterm.close_down ~howmany:n (type_of_cterm env' e') in
	     if not (ctype_equal defenv cdefenv e_t t) then failwith "a definition in letrec has wrong type!"
	 in
	 let _ = List.iter check_def defs in
	 let e_t = CbindCterm.close_down ~howmany:n (type_of_cterm env' (CbindCterm.open_up ~howmany:n e)) in
	   if ctype_equal defenv cdefenv (type_of_cterm env e_t) (CSort CType) then
	     e_t
	   else
	     failwith "letrec returning non-computational term")
    | CLet(v,d,e) ->
	(let _ = type_of ~case:v env d in
	 let e_t = type_of_cterm env (subst_bound ~case:v ~subst:d ~term:e) in
	   (* this was different: t was added to the env instead of being substituted *)
	   if ctype_equal defenv cdefenv (type_of_cterm env e_t) (CSort CType) then
	     e_t
	   else
	     failwith "let returning non-computational term")

    | CRefType(t) ->
	(let _ = check_computational env t in CSort(CType))
    | CMkRef(e,t') ->
	(let t = type_of_cterm env e in
	 let _ = check_computational env t in
	   if ctype_equal defenv cdefenv t t' then
	     CRefType(t)
	   else
	     failwith "designated ref type doesn't match the value"
	)
    | CAssign(e1,e2) ->
	(let t1 = ctype_fullwhdelta cdefenv (type_of_cterm env e1) in
	 let t2 = type_of_cterm env e2 in
	   match t1 with
	       CRefType(t) ->
		 if ctype_equal defenv cdefenv t2 t then
		   CUnitType
		 else
		   failwith "non type-compatible assignment"
	     | _ -> 
		 failwith "assignment to non-reference")
    | CReadRef(e) ->
	(let t = ctype_fullwhdelta cdefenv (type_of_cterm env e) in
	   match t with
	       CRefType(t) -> t
	     | _ -> failwith "dereferencing a non-reference")
    | CLoc(l,t) -> CRefType(t)
    | CSeq(e1,e2) ->
	(let t1 = type_of_cterm env e1 in
	 let t2 = type_of_cterm env e2 in
	 let _ = check_computational env t1 in
	 let _ = check_computational env t2 in
	   t2)

    | CIntType -> CSort(CType)
    | CIntConst(i) -> CIntType
    | CIntOp(o,e1,e2) ->
	(ensure_type_is env CIntType e1;
	 ensure_type_is env CIntType e2;
	 CIntType)
    | CIntTest(o,e1,e2) ->
	(ensure_type_is env CIntType e1;
	 ensure_type_is env CIntType e2;
	 CBoolType)
    | CBoolType -> CSort(CType)
    | CBoolConst(b) -> CBoolType
    | CBoolOp(o,e1,e2) ->
	(ensure_type_is env CBoolType e1;
	 ensure_type_is env CBoolType e2;
	 CBoolType)
    | CIfThenElse(e1,e2,e3) ->
	(ensure_type_is env CBoolType e1;
	 let t2 = type_of_cterm env e2 in
	 let t3 = type_of_cterm env e3 in
	   check_computational env t2;
	   if ctype_equal defenv cdefenv t2 t3 then
	     t2
	   else
	     failwith "types of if-branches do not match")

    | CArrayType(t) ->
	check_computational env t;
	CSort(CType)

    | CArrayLit(ts,t) ->
	(let _ = List.iter (fun e -> if not (ctype_equal defenv cdefenv (type_of_cterm env e) t) then failwith "array literal's members do not agree on type") ts in
	   CArrayType(t))

    | CMkArray(e1,e2,t_exp) ->
	  ensure_type_is env CIntType e1;
	  check_computational env t_exp;
	  ensure_type_is env t_exp e2;
	  CArrayType(t_exp)

    | CArrayGet(e1,e2) ->
	(let t1 = ctype_fullwhdelta cdefenv (type_of_cterm env e1) in
	  ensure_type_is env CIntType e2;
	  match t1 with
	      CArrayType(t) -> t
	    | _ -> failwith "array type expected when indexing array")

    | CArraySet(e1,e2,e3) ->
	(let t1 = ctype_fullwhdelta cdefenv (type_of_cterm env e1) in
	  ensure_type_is env CIntType e2;
	  match t1 with
	      CArrayType(t) ->
		if ctype_equal defenv cdefenv (type_of_cterm env e3) t then
		  CUnitType
		else
		  failwith "assigning different-type value to array index"
	    | _ -> failwith "array type expected when indexing array")

    | CArrayLen(e) ->
	(let t = ctype_fullwhdelta cdefenv (type_of_cterm env e) in
	  match t with
	      CArrayType(_) -> CIntType
	    | _ -> failwith "array type expected when using arraylen")

    | CArrayLoc(_,t) -> CArrayType(t)

    | CHolHash(e) ->
	(let t = type_of_cterm env e in
	   check_computational env t;
	   CIntType)

    | CPrint(e) ->
	(let t = type_of_cterm env e in
	   check_computational env t;
	   CUnitType)

    | CClosure(subst, e) ->
	(let (_,_,metaenv',ctxenv',compenv') as env' = env_from_subst defenv cdefenv subst in
	 let (<.>) = compose in
	 let e' = ((MetabindCterm.open_up ~howmany:(List.length metaenv')) <.>
	            (CtxbindCterm.open_up ~howmany:(List.length ctxenv')) <.>
		    (CbindCterm.open_up ~howmany:(List.length compenv'))) e in
	  type_of_cterm env' e')
	
and ensure_type_is ((defenv,cdefenv,_,_,_) as env) t e = 
  if not (ctype_equal defenv cdefenv (type_of_cterm env e) t) then failwith "expected another type"
    
and check_computational ((defenv,cdefenv,_,_,_) as env) t =

  if not (ctype_equal defenv cdefenv (type_of_cterm env t) (CSort(CType))) then failwith "computational type expected"

and check_pattern (defenv, _, _, _, _) metas t =

  (let ctx, tm = match t with CHolTerm(LTermInCtx(ctx,tm)) -> ctx, tm | _ -> failwith "patterns should be terms-in-context" in
     Logic_typing.check_pattern defenv ctx (List.length metas) tm)
	 
and check_branch ((defenv, cdefenv, metaenv, ctxenv, compenv) as env) typ ret_f (metas, t, e) =
  
  let open_up_meta n = MetabindCterm.open_up ~howmany:n in
  let env', n = List.fold_left (fun (env, n) (var, typ) ->
				  let typ' = open_up_meta n typ in
				  let _ = type_of_HOL env typ' in
				  let env' = add_to_env ~case:var typ' env in
				    (env', n+1)) (env,0) metas in
  let t' = open_up_meta n t in
  let t_typ = type_of_HOL env' t' in
  let _ = check_pattern env metas t in
  let e' = open_up_meta n e in
  let e_typ = type_of_cterm env' e' in
  let e_exp = open_up_meta n (ret_f t) in
    if ctype_equal defenv cdefenv t_typ typ then
      (if ctype_equal defenv cdefenv e_typ e_exp then
	 ()
       else
	 failwith "type of body of branch is not the expected one")
    else
      failwith "type of pattern is not the expected one"

and env_from_subst defenv cdefenv (metasubst,ctxsubst,compsubst) =
  let metaenv = List.map (type_of_modal ([], defenv, [], [])) metasubst in
  let ctxenv  = List.map (ctx_wf ([], defenv, [], [])) ctxsubst in
  let compenv = List.map (type_of_cterm (defenv, cdefenv, [], [], [])) compsubst in
    (defenv, cdefenv, metaenv, ctxenv, compenv)

and type_of_HOL (defenv, cdefenv, metaenv, ctxenv, compenv) (e : cterm) =
  match e with
      CHolTerm(t) ->
	(let tp = type_of_modal ([], defenv, metaenv, ctxenv) t in CHolTerm(tp))
    | _ -> failwith "non-HOL term in HOL position"
and csort_of_HOL (defenv, cdefenv, metaenv, ctxenv, compenv) (e : cterm) =
  match e with
      CHolTerm(LTermInCtx(_,_) as t) ->
	(let _ = type_of_modal ([], defenv, metaenv, ctxenv) t in CHol)
    | CHolTerm(_) -> failwith "metavariables are not allowed as types of metavariables"
    | _ -> failwith "non-HOL term in HOL abstractor position"
and type_of_ctx (defenv, cdefenv, metaenv, ctxenv, compenv) (e : cterm) =
  match e with
      CCtxTerm(t) -> (let _ = ctx_wf ([], defenv, metaenv, ctxenv) t in CSort(CCtx))
    | _ -> failwith "non-context term in context position"
and csort_of_ctx (defenv, cdefenv, metaenv, ctxenv, compenv) (e : cterm) =
  match e with
      CSort(CCtx) -> CCtx
    | _ -> failwith "non-context term in context abstractor position"
and csort_of_cterm ((defenv, cdefenv, metaenv, ctxenv, compenv) as env) (e : cterm) =
  match ctype_fullwhdelta cdefenv (type_of_cterm env e) with
      CSort(CType) -> CType
    | CSort(CKind) -> CKind
    | _ -> failwith "non-allowed classifier in computational abstractor position"
and csort_of ~case:s env t = case_kind csort_of_HOL csort_of_ctx csort_of_cterm ~case:s env t
and type_of  ~case:s env t = case_kind type_of_HOL type_of_ctx type_of_cterm ~case:s env t


let rec cterm_def_add name tm ?(expected_tp=None) defenv cdefenv =
  let tp = type_of_cterm (defenv,cdefenv,[],[],[]) tm in
  let add name what cdefenv =
    (if Dict.mem name cdefenv then
       failwith ("definition " ^ name ^ " already exists")
     else
       Dict.add name what cdefenv)
  in
    match expected_tp with
	Some(t) ->
	  (if ctype_equal defenv cdefenv tp t then
	     add name (t, tm) cdefenv
	   else
	     failwith "the term's type doesn't match the definition type")
      | None ->
	  add name (tp, tm) cdefenv

let comp_global_env : cterm_defenv ref = ref Dict.empty
let comp_global_reset () =
  let (d1,d2) = !Logic_main.logic_global_env in
    comp_global_env := Dict.empty;
    Logic_main.logic_global_env := (d1, Dict.empty)

let comp_gettype tm =
  let case = match tm with CHolTerm(_) -> (None, CHol) | CCtxTerm(_) -> (None, CCtx) | _ -> (None, CType) in
    type_of ~case:case (!Logic_main.logic_global_env, !comp_global_env, [], [], []) tm
let comp_print tm =
  match tm with
      CNVar(s) -> Comp_print.print_cterm (get_cdef_term s !comp_global_env)
    | _ -> Comp_print.print_cterm tm
let comp_print_type tm =
  Comp_print.print_cterm (comp_gettype tm)
